# -*- coding: utf-8 -*-
"""
Plotting function for visualizing comparison results.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import warnings
import numpy as np
import math
import scipy.cluster.hierarchy as shc
import scipy.spatial.distance as ssd
from sklearn.manifold import TSNE

import matplotlib.pyplot as plt
import matplotlib as mpl
from nilearn.plotting import plot_markers
from matplotlib.colors import ListedColormap
import seaborn as sns
import pandas as pd
import os

from ..dfc_utils import visualize_conn_mat_dict

################################# Parameters ####################################

fig_dpi = 120
fig_bbox_inches = 'tight'
fig_pad = 0.1
show_title = False
save_fig_format = 'png' # pdf, png, 

################################# Plotting Functions ####################################

def title2file_name(title):
    '''
    change all spaces in the title to _
    the original string remains unchanged 
    '''
    return title.replace(" ", "_")


def plot_TS(X, Fs, title=''):
    '''
    X = (n_time, n_features)
    '''
    if X.shape[1] > 10:
        print('Too many features to plot')
        return
    fig_width = 35
    fig_height = 2 * X.shape[1]
    fig, axes = plt.subplots(X.shape[1], 1, figsize=(fig_width, fig_height),
        facecolor='w', edgecolor='k')
    time = np.arange(0, X.shape[0])/Fs
    for i in range(0, X.shape[1]):
        axes[i].plot(time, X[:, i], linewidth=4)
        axes[i].set_title('TC'+str(i), fontdict={'fontsize': 15, 'fontweight': 'bold'})
    plt.suptitle(title)
    plt.xlabel('Time (s)')
    plt.show()
    # if save_image:
    #     folder = output_root[:output_root.rfind('/')]
    #     if not os.path.exists(folder):
    #         os.makedirs(folder)
    #     fig.savefig(output_root+title.replace(" ", "_")+'.'+save_fig_format, 
    #         dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
    #     ) 
    #     plt.close()
    # else:
    #     plt.show()

def plot_sample_dFC(D, x,
    title='',
    cmap='seismic',
    normalize=False,
    disp_diag=True,
    save_image=False, output_root=None, 
    fix_lim=True, center_0=True, 
    node_networks=None, segmented=False 
    ):
    '''
    D is a dictionary of dFC samples. each
    key is the name of a dFC matrix (e.g. method
    used for assessing it), and D[key][x] contains the 
    the dFC matrix as a numpy ndarray 
    '''

    num_dFC = len(D)
    names_lst = [key for key in D]
    num_time = len(D[names_lst[0]][x])

    fig_width = 48*(num_time/10)
    fig_height = 55*(num_dFC/10)

    fig, axes = plt.subplots(num_dFC, num_time, figsize=(fig_width, fig_height), \
        facecolor='w', edgecolor='k')

    fig.subplots_adjust(
        bottom=0.1,
        top=0.85,
        left=0.1,
        right=0.9,
        wspace=0.5,
        hspace=0.6
    )

    for i, dFC_mat_name in enumerate(D):
        visualize_conn_mat_dict(data=D[dFC_mat_name][x], 
            node_networks=node_networks, 
            title=dFC_mat_name, 
            cmap=cmap, center_0=center_0,
            normalize=normalize, fix_lim=fix_lim, 
            disp_diag=disp_diag,
            segmented=segmented,
            save_image=False, output_root=output_root,
            axes=axes[i, :], fig=fig, 
        )

    fig.subplots_adjust(
        bottom=0.1,
        top=0.85,
        left=0.1,
        right=0.9,
        wspace=0.5,
        hspace=0.6
    )
        
    # set row names
    for i, dFC_mat_name in enumerate(D):
        axes[i, 0].set_ylabel(dFC_mat_name, fontdict={'fontsize': 25, 'fontweight': 'bold'}, rotation=90)
    
    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        fig.savefig(output_root+title.replace(" ", "_")+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()
    else:
        plt.show()

    
def pairwise_cat_plots(data=None, x=None, y=None, z=None,
    title='', 
    label_dict={},
    save_image=False, output_root=None
    ):
    '''
    data is a dictionary with different vars as keys 
    if z is specidied, it will be used as a out of distribution 
    sample, e.g. actual similarity when plotting randomized 
    distribution.
    '''

    sns.set_context("paper", 
        font_scale=2.5, 
        rc={"lines.linewidth": 3.0}
    )

    row_keys = [key for key in data]
    n_rows = len(row_keys)
    column_keys = [key for key in data[row_keys[-1]]]
    n_columns = len(column_keys)

    sns.set_style('darkgrid')

    fig_width = n_columns * 5
    fig_height = n_rows * 5
    fig, axs = plt.subplots(n_rows, n_columns, figsize=(fig_width, fig_height), \
        facecolor='w', edgecolor='k', sharex=True, sharey=True)
    
    axs_plotted = list()
    for i, key_i in enumerate(data):
        for j, key_j in enumerate(data[key_i]):
            df = pd.DataFrame(data[key_i][key_j])

            if not z is None:
                sns.stripplot(ax=axs[i, j], data=df, x=x, y=z, color='red', jitter=False, size=10)
            sns.violinplot(ax=axs[i, j], data=df, x=x, y=y)

            axs[i, j].set_title(key_i+'-'+key_j, fontdict={'fontsize': 25, 'fontweight': 'bold'})

            ## set labels
            ylabel = axs[i, j].get_ylabel()
            if ylabel in label_dict:
                ylabel = label_dict[ylabel]
            axs[i, j].set_ylabel(ylabel, fontdict={'fontsize': 20, 'fontweight': 'bold'})
            xlabel = axs[i, j].get_xlabel()
            if xlabel in label_dict:
                xlabel = label_dict[xlabel]
            axs[i, j].set_xlabel(xlabel, fontdict={'fontsize': 20, 'fontweight': 'bold'})
            # set font size of the tick labels and make them bold 
            tick_labels = axs[i, j].get_xticklabels() + axs[i, j].get_yticklabels()
            for label in tick_labels:
                label.set_fontweight('bold')

            axs_plotted.append(axs[i, j])

    # remove extra subplots
    for ax in axs.ravel():
        if not ax in axs_plotted:
            ax.set_axis_off()
            ax.xaxis.set_tick_params(which='both', labelbottom=True)
    
    if show_title:
        plt.suptitle(title, fontsize=15, y=0.90)

    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title2file_name(title)+'.'+save_fig_format, \
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format \
        ) 
        plt.close()
    else:
        plt.show()

def joint_dist_plot(data,
    title='',
    label_dict={},
    save_image=False, output_root=None
    ):
    '''
    data is a dictionary including list of dFC values
    of each dFC method
    '''
    df = pd.DataFrame(data)
    fig_width = 5*len(data)
    fig_height = 5*len(data)

    sns.set_context("paper", 
        font_scale=2.5, 
        rc={"lines.linewidth": 3.0}
    )
    
    sns.set_style('darkgrid')

    g = sns.PairGrid(df)

    g.map_diag(sns.histplot)
    g.map_offdiag(sns.histplot)

    g.fig.set_figwidth(fig_width)
    g.fig.set_figheight(fig_height)
    g.fig.subplots_adjust(top=0.95)

    for i in range(g.axes.shape[0]):
        for j in range(g.axes.shape[1]):

            ## set labels
            ylabel = g.axes[i, j].get_ylabel()
            if ylabel in label_dict:
                ylabel = label_dict[ylabel]
            g.axes[i, j].set_ylabel(ylabel, fontdict={'fontsize': 25, 'fontweight': 'bold'})
            xlabel = g.axes[i, j].get_xlabel()
            if xlabel in label_dict:
                xlabel = label_dict[xlabel]
            g.axes[i, j].set_xlabel(xlabel, fontdict={'fontsize': 25, 'fontweight': 'bold'})

            # set font size of the tick labels and make them bold 
            tick_labels = g.axes[i, j].get_xticklabels() + g.axes[i, j].get_yticklabels()
            for label in tick_labels:
                label.set_fontweight('bold')

    if show_title:
        plt.suptitle(title, fontsize=50, y=0.98)

    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title2file_name(title)+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()
    else:
        plt.show()

def pairwise_scatter_plots(data, x, y,
    title='', hist=False,
    label_dict={},
    equal_axis_lim=False, show_x_equal_y=False,
    save_image=False, output_root=None
    ):
    '''
    data is a dictionary with different vars as keys 
    '''

    sns.set_context("paper", 
        font_scale=2.5, 
        rc={"lines.linewidth": 3.0}
    )

    row_keys = [key for key in data]
    n_rows = len(row_keys)
    column_keys = [key for key in data[row_keys[-1]]]
    n_columns = len(column_keys)

    sns.set_style('darkgrid')

    fig_width = n_columns * 5
    fig_height = n_rows * 5
    fig, axs = plt.subplots(n_rows, n_columns, figsize=(fig_width, fig_height), \
        facecolor='w', edgecolor='k', sharex=True, sharey=True)
    
    # equal x_lim and y_lim
    if equal_axis_lim or show_x_equal_y:
        min_lim = None
        max_lim = None
        for i, key_i in enumerate(data):
            for j, key_j in enumerate(data[key_i]):
                df = pd.DataFrame(data[key_i][key_j])
                m = np.minimum(df[x].min(), df[y].min())
                M = np.maximum(df[x].max(), df[y].max())
                if min_lim is None:
                    min_lim = m
                    max_lim = M
                else:
                    min_lim = np.minimum(m, min_lim)
                    max_lim = np.maximum(M, max_lim)
                
        lim_L = max_lim - min_lim
        min_lim = min_lim - lim_L*0.1
        max_lim = max_lim + lim_L*0.1

    axs_plotted = list()
    for i, key_i in enumerate(data):
        for j, key_j in enumerate(data[key_i]):
            df = pd.DataFrame(data[key_i][key_j])
            if hist:
                g = sns.histplot(ax=axs[i, j], data=df, x=x, y=y, bins=50)
            else:
                g = sns.scatterplot(ax=axs[i, j], data=df, x=x, y=y, s=50)
            axs[i, j].set_title(key_i+'-'+key_j, fontdict={'fontsize': 25, 'fontweight': 'bold'})

            ## set labels and font sizes
            ylabel = g.get_ylabel()
            if ylabel in label_dict:
                ylabel = label_dict[ylabel]
            g.set_ylabel(ylabel, fontdict={'fontsize': 18, 'fontweight': 'bold'})
            xlabel = g.get_xlabel()
            if xlabel in label_dict:
                xlabel = label_dict[xlabel]
            g.set_xlabel(xlabel, fontdict={'fontsize': 18, 'fontweight': 'bold'})
            g.tick_params(axis='x', which='major', labelsize=18)
            g.tick_params(axis='y', which='major', labelsize=18)   
            tick_labels = g.get_xticklabels() + g.get_yticklabels()
            for label in tick_labels:
                label.set_fontweight('bold')

            # equal x_lim and y_lim
            if equal_axis_lim:
                axs[i, j].set_xlim(min_lim, max_lim)
                axs[i, j].set_ylim(min_lim, max_lim)

            # y=x line
            if show_x_equal_y:
                X_plot = np.linspace(min_lim, max_lim, 100)
                Y_plot = X_plot
                axs[i, j].plot(X_plot, Y_plot, color='r')

            axs_plotted.append(axs[i, j])
    # remove extra subplots
    for ax in axs.ravel():
        if not ax in axs_plotted:
            ax.set_axis_off()
            ax.xaxis.set_tick_params(which='both', labelbottom=True)
    
    if show_title:
        plt.suptitle(title, fontsize=15, y=0.90)

    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title2file_name(title)+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format 
        ) 
        plt.close()
    else:
        plt.show()

def scatter_plot(data, x, y,
    labels=None, hue=None,
    title='', hist=False,
    label_dict={},
    equal_axis_lim=False, show_x_equal_y=False,
    c=0.25,
    save_image=False, output_root=None
    ):
    '''
    data is a dictionary with different vars as keys 
    c determines how far the annotation will be from dots
    '''
    df = pd.DataFrame(data)

    sns.set_context("paper", 
        font_scale=2.5, 
        rc={"lines.linewidth": 3.0}
    )

    fig_width = 20
    fig_height = 20 
    plt.figure(figsize=(fig_width, fig_height))
    sns.set_style('darkgrid')
    if hist:
        g = sns.histplot(data=df, x=x, y=y, hue=hue)
    else:
        g = sns.scatterplot(data=df, x=x, y=y, s=100, hue=hue)

    ## set labels and font sizes
    ylabel = g.get_ylabel()
    if ylabel in label_dict:
        ylabel = label_dict[ylabel]
    g.set_ylabel(ylabel, fontdict={'fontsize': 35, 'fontweight': 'bold'})
    xlabel = g.get_xlabel()
    if xlabel in label_dict:
        xlabel = label_dict[xlabel]
    g.set_xlabel(xlabel, fontdict={'fontsize': 35, 'fontweight': 'bold'})
    g.tick_params(axis='x', which='major', labelsize=30)
    g.tick_params(axis='y', which='major', labelsize=30)   
    tick_labels = g.get_xticklabels() + g.get_yticklabels()
    for label in tick_labels:
        label.set_fontweight('bold')
    
    # equal x_lim and y_lim
    if equal_axis_lim:
        min_lim = np.minimum(df[x].min(), df[y].min())
        max_lim = np.maximum(df[x].max(), df[y].max())
        lim_L = max_lim - min_lim
        min_lim = min_lim - lim_L*0.1
        max_lim = max_lim + lim_L*0.1
        g.set_xlim(min_lim, max_lim)
        g.set_ylim(min_lim, max_lim)

    # y=x line
    if show_x_equal_y:
        min_lim = np.minimum(df[x].min(), df[y].min())
        max_lim = np.maximum(df[x].max(), df[y].max())
        lim_L = max_lim - min_lim
        min_lim = min_lim - lim_L*0.1
        max_lim = max_lim + lim_L*0.1
        X_plot = np.linspace(min_lim, max_lim, 100)
        Y_plot = X_plot
        plt.plot(X_plot, Y_plot, color='r')

    if (not labels is None) and (not hist):
        # the labels are located smartly
        # the direction will be away from mean
        # the distance will be inverse proportional to 
        # distance from mean
        mid_x = (np.max(df[x]) + np.min(df[x]))/2
        mid_y = (np.max(df[y]) + np.min(df[y]))/2
        x_range = max(np.max(df[x])-mid_x, mid_x-np.min(df[x]))
        y_range = max(np.max(df[y])-mid_y, mid_y-np.min(df[y]))
        distance_from_mean_range = math.sqrt(x_range**2+y_range**2)
        for i in range(len(df[x])):
            distance_from_mean = math.sqrt((df[x][i]-mid_x)**2+(df[y][i]-mid_y)**2)
            text_x = df[x][i]+c*np.sign(df[x][i]-mid_x)*np.abs(df[x][i]-mid_x)*(distance_from_mean_range-distance_from_mean)/distance_from_mean 
            text_y = df[y][i]+c*np.sign(df[y][i]-mid_y)*np.abs(df[y][i]-mid_y)*(distance_from_mean_range-distance_from_mean)/distance_from_mean
            plt.text(
                x=text_x, 
                y=text_y, 
                s=df[labels][i], 
                fontdict=dict(color='black', size=14, weight='bold'),
            )
            plt.plot(
                [df[x][i], text_x], [df[y][i], text_y], 
                'k', linewidth=0.5
            )
    
    if show_title:
        plt.title(title, fontsize=15)

    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title2file_name(title)+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()
    else:
        plt.show()

def cat_plot(data, x, y, 
    kind='bar',
    scale_dist=False,
    log=False,
    title='',
    label_dict={},
    y_lim=None,
    save_image=False, output_root=None
    ):
    '''
    data is a dictionary with different vars as keys 
    kind can be = box or violin or bar
    scale_dist is only for kind=='violin'
    '''

    sns.set_context("paper", 
        font_scale=1.0, 
        rc={"lines.linewidth": 1.0}
    )

    sns.set_style('darkgrid')

    df = pd.DataFrame(data)

    fig_width = 2*len(np.unique(data[x]))
    fig_height = 5 

    if kind=='violin' and scale_dist:
        g = sns.catplot(data=df, x=x, y=y, kind=kind,
            scale='width'
            # errorbar=("pi", 95)
        )
    elif kind=='bar':
        g = sns.catplot(data=df, x=x, y=y, kind=kind,
                        width=0.25
            # errorbar=("pi", 95)
        )
    elif kind=='box':
        g = sns.catplot(data=df, x=x, y=y, kind=kind,
                        width=0.25, fliersize=1.0
            # errorbar=("pi", 95)
        )
    else:
        g = sns.catplot(data=df, x=x, y=y, kind=kind,
            # errorbar=("pi", 95)
        )

    if log:
        plt.yscale('log')

    g.fig.set_figwidth(fig_width)
    g.fig.set_figheight(fig_height)

    ## set labels
    ylabel = g.ax.get_ylabel()
    if ylabel in label_dict:
        ylabel = label_dict[ylabel]
    g.ax.set_ylabel(ylabel, fontdict={'fontsize': 13, 'fontweight': 'bold'})
    xlabel = g.ax.get_xlabel()
    if xlabel in label_dict:
        xlabel = label_dict[xlabel]
    g.ax.set_xlabel(xlabel, fontdict={'fontsize': 13, 'fontweight': 'bold'})
    # set font size of the tick labels and make them bold
    tick_labels = g.ax.get_xticklabels() + g.ax.get_yticklabels()
    for label in tick_labels:
        label.set_fontweight('bold')
    if not y_lim is None:
        g.ax.set_ylim(y_lim)

    if show_title:
        plt.title(title, fontsize=15)
    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title2file_name(title)+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()
    else:
        plt.show()

def visualize_sim_mat(data, mat_key, title='', 
    name_lst_key=None, 
    cmap='viridis',
    annot=True, fmt=2, 
    label_dict={},
    show_diag=False, show_sig=False, no_color=False,
    save_image=False, output_root=None, axes=None, fig=None, 
    ):

    '''
    - name_lst_key is the key to list of names
    - data must be a dict of correlation/connectivity matrices
    - masks the nan values
    sample:
    Suptitle1
        corr_mat
            0.00 0.31 0.76 
            0.31 0.00 0.43 
            0.76 0.43 0.00 
        measure_lst
            ContinuousHMM
            Windowless
            Clustering_pear_corr
    Suptitle1
        corr_mat
            0.00 0.32 0.76 
            0.32 0.00 0.45 
            0.76 0.45 0.00 
        measure_lst
            ContinuousHMM
            Windowless
            Clustering_pear_corr
    '''

    sns.set_context("paper", 
        font_scale=1.0, 
        rc={"lines.linewidth": 1.0}
    )

    sns.set_style('white')

    if no_color:
        cmap = ListedColormap(['white'])
    
    if name_lst_key is None:
        fig_width = int(25*(len(data)/10))
    else:
        fig_width = int(60*(len(data)/10) + 1)
    fig_height = 5 

    fig_flag = True
    if axes is None or fig is None:
        fig_flag = False

    if not fig_flag:
        fig, axes = plt.subplots(1, len(data), figsize=(fig_width, fig_height),
            facecolor='w', edgecolor='k', sharey=False
        )

    if not type(axes) is np.ndarray:
        axes = np.array([axes])

    if show_title:
        fig.suptitle(title, fontsize=20, y=0.98) #, fontsize=20, size=20

    axes = axes.ravel()

    # normalizing and scale
    sim_mats = list()
    for i, key in enumerate(data):
        sim_mats.append(data[key][mat_key])
    sim_mats = np.array(sim_mats)

    # plot
    for i, key in enumerate(data):

        C = sim_mats[i,:,:]

        name_lst = None
        if not name_lst_key is None:
            name_lst = data[key][name_lst_key]

        cbar_flag = False
        # if i==(len(data)-1):
        #     cbar_flag = True

        if annot:
            C_forlabels = C.copy()
            if not show_diag:
                np.fill_diagonal(C_forlabels, np.nan)
            df = pd.DataFrame(C_forlabels)
            if show_sig:
                annot_labels = df.applymap(lambda v: '' if np.isnan(v) else str(round(v, fmt))+''.join(['*' for t in [.05, .01, .001] if v<=t]))
            else:
                annot_labels = df.applymap(lambda v: '' if np.isnan(v) else str(round(v, fmt)))
        else:
            annot_labels = False

        # borderlines color
        if no_color:
            linecolor = 'black'
            annot_kws={'weight': 'bold'}
        else:
            linecolor = 'w'
            annot_kws={'weight': 'bold'}

        im = sns.heatmap(C, 
            annot=annot_labels, annot_kws=annot_kws,
            fmt='', cmap=cmap, 
            xticklabels=name_lst, yticklabels=name_lst, 
            ax=axes[i], cbar=cbar_flag,
            square=True, linewidth=2, linecolor=linecolor
        )
        axis_title = key
        if key in label_dict:
            axis_title = label_dict[key]
        axes[i].set_title(axis_title, fontdict= {'fontsize': 18, 'fontweight':'bold'})
        im.set_xticklabels(im.get_xticklabels(), fontdict= {'fontsize': 14, 'fontweight':'bold'}, rotation=90)
        im.set_yticklabels(im.get_yticklabels(), fontdict= {'fontsize': 14, 'fontweight':'bold'}, rotation=0)

    if not fig_flag:
            
        fig.subplots_adjust(
            bottom=0.1, 
            top=0.85, 
            left=0.1, 
            right=0.9,
        )

        if not name_lst is None:
            fig.subplots_adjust(
                wspace=0.5
            )

    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title2file_name(title)+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()
    else:
        plt.show()

def distance2Z(dist_mat, method='ward'):
    # convert the redundant n*n square matrix form into a condensed nC2 array
    distArray = ssd.squareform(dist_mat) 
    Z = shc.linkage(distArray, method=method)
    return Z

def dist_mat_dendo(Z, labels, 
    distances_CI=None,
    title='', \
    save_image=False, output_root=None, \
    ):
    '''
    if distances_CI is provided, confidence intervals (CI)
    of the distances will be shown. the order should be the same as 
    Z
    '''

    sns.set_context("paper", 
        font_scale=3.5, 
        rc={"lines.linewidth": 3.0,
            'font.weight': 'bold'
            }
    )

    sns.set_style('darkgrid')

    width = int(2.5*len(Z))
    fig = plt.figure(figsize=(width, 5))
    ax = fig.add_subplot(1, 1, 1)    
    with mpl.rc_context({'lines.linewidth': 3}):

        dend = shc.dendrogram(Z, distance_sort='ascending', no_plot=False, labels=labels)

        # show confidence interval of distances
        if not distances_CI is None:

            max_y_lim = None
            for i, d in zip(dend['icoord'], dend['dcoord']):
                
                # we have to match the distances in dcoord
                # with those in Z, because the orders are not
                # the same
                count = 0
                for idx, clstr in enumerate(Z):
                    if np.isclose(Z[idx][2], d[1]):
                        count += 1
                        Z_CI = distances_CI[idx]

                if count > 1 or count==0:
                    warnings.warn(
                        'Error in finding std of linkage.',
                        UserWarning
                    )

                x = 0.5 * sum(i[1:3])
                y = d[1]
                ci_line_y = np.linspace(y-Z_CI, y+Z_CI, 100)
                ci_line_x = x * np.ones(100)
                # cut start and the end for better
                # visualization
                ci_line_y = ci_line_y[5:-5]
                ci_line_x = ci_line_x[5:-5]

                plt.plot(ci_line_x, ci_line_y, 'black')
                plt.plot(x, y-Z_CI, 'k_', markersize=15, linewidth=15)
                plt.plot(x, y+Z_CI, 'k_', markersize=15, linewidth=15)
                plt.plot(x, y, 'wo', markersize=5, mec='k')
                # plt.annotate("%.2g" % y, (x, y), xytext=(15, 13),
                #             fontsize = 11,
                #             fontweight= 'bold',
                #             textcoords='offset points',
                #             va='top', ha='center')
                if max_y_lim is None:
                    max_y_lim = y+Z_CI
                else:
                    max_y_lim = max(y+Z_CI, max_y_lim)
            plt.ylim(0, max_y_lim*1.1)
                
    if show_title:
        plt.title(title, fontsize=15)
        
    # set font size of the tick labels and make them bold
    ax.tick_params(axis='x', which='major', labelsize=15)
    ax.tick_params(axis='y', which='major', labelsize=15)   
    tick_labels = ax.get_xticklabels() + ax.get_yticklabels()
    for label in tick_labels:
        label.set_fontweight('bold')

    # save figure
    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title2file_name(title)+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()

def plot_TSNE(
        dist_mat, 
        sample_measure_lst,
        color_dict,
        projection='2d',
        title='',
        save_image=False, output_root=None,
    ):

    sns.set_context("paper", 
        font_scale=2.5, 
        rc={
            "lines.linewidth": 3.0,
            "lines.markersize": 10.0
            }
    )

    sns.set_style('darkgrid')

    fig_width = 20
    fig_height = 20 
    
    if projection=='2d':
        X_embedded = TSNE(
                        n_components=2, 
                        learning_rate='auto',
                        init='random', perplexity=30, 
                        metric='precomputed'
                    ).fit_transform(dist_mat)

        # 2D plot
        plt.figure(figsize=(fig_width, fig_height))
        sns.scatterplot(
            x=X_embedded[:, 0], y=X_embedded[:, 1], 
            hue=sample_measure_lst, 
            palette=color_dict,
            alpha=0.7
        )
    elif projection=='3d':
        X_embedded = TSNE(
                        n_components=3, 
                        learning_rate='auto',
                        init='random', perplexity=30, 
                        metric='precomputed'
                    ).fit_transform(dist_mat)
        
        measures_lst = list(set(sample_measure_lst))
        measures_lst.sort()

        # 3D plot
        fig = plt.figure(figsize=(fig_width, fig_height))
        ax = fig.add_subplot(projection='3d')
        sample_measure_array = np.array(sample_measure_lst)
        for measure in measures_lst:
            scatter = ax.scatter(
                X_embedded[sample_measure_array==measure, 0], 
                X_embedded[sample_measure_array==measure, 1], 
                X_embedded[sample_measure_array==measure, 2],
                c=color_dict[measure],
                label=measure
            )
        ax.legend()

    if show_title:
        plt.title(title, fontsize=15)

    # save figure
    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title2file_name(title)+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()

def plot_brain_act(act_vec, locs, axes,
    title='', save_image=False, output_root=''
    ):

    plot_markers(
        node_values=act_vec, node_coords=locs, 
        node_cmap='hot', 
        display_mode='z', 
        colorbar=False, axes=axes
    )

    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title2file_name(title)+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()

def visualize_state_TC(TC_lst, \
    TRs, \
    state_lst, \
    TC_name_lst, \
    title='', \
    save_image=None, output_root=None\
    ):

    color_lst = ['k', 'b', 'g', 'r']

    if 'on' in state_lst and 'off' in state_lst:
        ticks = range(2)
    else:
        ticks = range(1, len(state_lst)+1)

    plt.figure(figsize=(25, 5))
    for i, TC in enumerate(TC_lst):
        plt.plot(TRs, TC, color_lst[i], linewidth=2)
    plt.xlabel('TR')
    plt.yticks(ticks=ticks, labels=state_lst)
    plt.legend(TC_name_lst)
    if show_title:
        plt.title(title)
    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title2file_name(title)+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()
    # else:
    #     plt.show()

    return



"""
Implementation of dFC methods.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import numpy as np
import h5py
import os

from .dfc_utils import intersection, label2network
from .time_series import TIME_SERIES

################################# DATA_LOADER functions ######################################

def find_subj_list(data_root, sessions):
    '''
    find the list of subjects in data_root
    the files must follow the format: subjectID_sessionID
    only these files should be in the data_root
    '''
    if data_root[-1] != '/':
        data_root += '/'

    ALL_FILES = os.listdir(data_root)
    FOLDERS = [item for item in ALL_FILES if os.path.isdir(data_root+item)]
    
    FOLDERS.sort()
    SUBJECTS = list()
    for s in FOLDERS:
        num = s[:s.find('_')]
        SUBJECTS.append(num)
    # the subjects might be repeated because of different sessions
    SUBJECTS = list(set(SUBJECTS))
    SUBJECTS.sort()

    print( str(len(SUBJECTS)) + ' subjects were found. ')

    failed_subjs = []
    kept_subjs = []
    for subj in SUBJECTS:
        kept_subjs.append(subj)
        for session in sessions:
            if not os.path.exists(f"{data_root}{subj}_{session}"):
                failed_subjs.append(subj)
                kept_subjs.remove(subj)
                break

    print( str(len(failed_subjs)) + ' subjects had missing sessions. ' + str(len(kept_subjs)) + ' subjects were kept. ')

    return kept_subjs

def load_from_array(subj_id2load=None, **params):
    '''
    load fMRI data from numpy or mat files
    input time_series.shape must be (time, roi)
    returns a dictionary of TIME_SERIES objects
    each corresponding to a session

    - if the file_name is a .mat file, it will be loaded using h5py
      if the file_name is a .npy file, it will be loaded using np.load

    - the roi locations should be in the same folder and a .npy file
      with the name: params['roi_locs_file']
      it must 

    - and the roi labels should be in the same folder and a .npy file
      with the name: params['roi_labels_file']
      it must be a list of strings

    - labels should be in the format: Hemisphere_Network_ID
        ow, the network2include will not work properly
    '''

    SESSIONs = params['SESSIONs'] # list of sessions
    if subj_id2load is None:
        SUBJECTS = find_subj_list(params['data_root'], sessions=SESSIONs)
    else:
        SUBJECTS = [subj_id2load]

    # LOAD Region Location DATA
    locs = np.load(params['data_root']+params['roi_locs_file'], allow_pickle='True').item()
    locs = locs['locs']

    # LOAD Region Labels DATA
    labels = np.load(params['data_root']+params['roi_labels_file'], allow_pickle='True').item()
    labels = labels['labels']

    assert type(locs) is np.ndarray, 'locs must be a numpy array'
    assert type(labels) is list, 'labels must be a list'
    assert locs.shape[0] == len(labels), 'locs and labels must have the same length'
    assert locs.shape[1] == 3, 'locs must have 3 columns'

    # apply networks2include
    # if params['networks2include'] is None, all the regions will be included
    if not params['networks2include'] is None:
        nodes2include = [i for i, x in enumerate(labels) if label2network(x) in params['networks2include']]
    else:
        nodes2include = [i for i, x in enumerate(labels)]
    locs = locs[nodes2include, :]
    labels = [x for node, x in enumerate(labels) if node in nodes2include]


    BOLD = {}
    for session in SESSIONs:
        BOLD[session] = None
        for subject in SUBJECTS:

            subj_fldr = subject + '_' + session

            # LOAD BOLD Data

            if params['file_name'][params['file_name'].find('.'):] == '.mat':
                with h5py.File(params['data_root']+subj_fldr+'/'+params['file_name'], 'r') as f:
                    DATA = {k: np.array(f[k]) for k in f.keys()}
            elif params['file_name'][params['file_name'].find('.'):] == '.npy':
                DATA = np.load(params['data_root']+subj_fldr+'/'+params['file_name'], allow_pickle='True').item()
            time_series = DATA['ROI_data'] # time_series.shape = (time, roi)

            # change time_series.shape to (roi, time)
            time_series = time_series.T

            # apply networks2include
            time_series = time_series[nodes2include, :]

            if BOLD[session] is None:
                BOLD[session] = TIME_SERIES(data=time_series, subj_id=subject,
                                    Fs=params['Fs'],
                                    locs=locs, node_labels=labels,
                                    TS_name='BOLD Real', session_name=session
                                )
            else:
                BOLD[session].append_ts(new_time_series=time_series, subj_id=subject)

        print( '*** Session ' + session + ': ' )
        print( 'number of regions= '+str(BOLD[session].n_regions) + ', number of time points= ' + str(BOLD[session].n_time) )

    return BOLD


def nifti2array(nifti_file, 
                confound_strategy='none', standardize=False,
                n_rois=100
    ):
    '''
    this function uses nilearn maskers to extract 
    BOLD signals from nifti files
    For now it only works with schaefer atlas,
    but you can set the number of rois to extract
    {100, 200, 300, 400, 500, 600, 700, 800, 900, 1000}

    returns a numpy array of shape (time, roi)
    and labels and locs of rois

    confound_strategy:
        'none': no confounds are used
        'no_motion': motion parameters are used
        'no_motion_no_gsr': motion parameters are used 
                            and global signal regression 
                            is applied.
    '''
    from nilearn.maskers import NiftiLabelsMasker
    from nilearn import datasets
    from nilearn.plotting import find_parcellation_cut_coords
    from nilearn.interfaces.fmriprep import load_confounds

    parc = datasets.fetch_atlas_schaefer_2018(n_rois=n_rois)
    atlas_filename = parc.maps
    labels = parc.labels
    # The list of labels does not contain ‘Background’ by default. 
    # To have proper indexing, you should either manually add ‘Background’ to the list of labels:
    # Prepend background label
    labels = np.insert(labels, 0, 'Background')

    # extract locs
    # test!
    # check if order is the same as labels
    locs, labels_ = find_parcellation_cut_coords(
                        atlas_filename, 
                        background_label=0, 
                        return_label_names=True
                    )

    # create the masker for extracting time series
    masker = NiftiLabelsMasker(
        labels_img=atlas_filename, 
        labels=labels, 
        resampling_target='data',
        standardize=standardize,
    )
    
    labels = np.delete(labels, 0) # remove the background label
    labels = [label.decode() for label in labels]
    
    ### extract the timeseries
    if confound_strategy=='none':
        time_series = masker.fit_transform(nifti_file)
    elif confound_strategy=='no_motion':
        confounds_simple, sample_mask = load_confounds(
            nifti_file,
            strategy=["high_pass", "motion", "wm_csf"],
            motion="basic", wm_csf="basic"
        )
        time_series = masker.fit_transform(
            nifti_file,
            confounds=confounds_simple,
            sample_mask=sample_mask
        )
    elif confound_strategy=='no_motion_no_gsr':
        confounds_simple, sample_mask = load_confounds(
            nifti_file,
            strategy=["high_pass", "motion", "wm_csf", "global_signal"],
            motion="basic", wm_csf="basic", global_signal="basic"
        )
        time_series = masker.fit_transform(
            nifti_file,
            confounds=confounds_simple,
            sample_mask=sample_mask
        )

    return time_series, labels, locs


def nifti2timeseries(
        nifti_file, 
        n_rois, Fs,
        subj_id,
        confound_strategy='none', 
        standardize=False,
        TS_name=None,
        session=None,
    ):
    '''
    this function is only for single subject and single session data loading 
    it uses nilearn maskers to extract ROI signals from nifti files
    and returns a TIME_SERIES object

    For now it only works with schaefer atlas,
    but you can set the number of rois to extract
    {100, 200, 300, 400, 500, 600, 700, 800, 900, 1000}
    '''
    time_series, labels, locs = nifti2array(
        nifti_file=nifti_file, 
        confound_strategy=confound_strategy, 
        standardize=standardize,
        n_rois=n_rois
    )

    assert type(locs) is np.ndarray, 'locs must be a numpy array'
    assert type(labels) is list, 'labels must be a list'
    assert locs.shape[0] == len(labels), 'locs and labels must have the same length'
    assert locs.shape[1] == 3, 'locs must have 3 columns'

    # change time_series.shape to (roi, time)
    time_series = time_series.T

    if TS_name is None:
        TS_name = subj_id + ' time series'
    
    BOLD = TIME_SERIES(
                data=time_series, subj_id=subj_id,
                Fs=Fs,
                locs=locs, node_labels=labels,
                TS_name=TS_name, session_name=session
    )

    return BOLD


def multi_nifti2timeseries(
        nifti_files_list,
        subj_id_list,
        n_rois, Fs,
        confound_strategy='none', 
        standardize=False,
        TS_name=None,
        session=None,
):
    '''
    loading data of multiple subjects, but single session, from their nifti files
    '''
    BOLD_multi = None
    for subj_id, nifti_file in zip(subj_id_list, nifti_files_list):
        if BOLD_multi is None:
            BOLD_multi = nifti2timeseries(
                            nifti_file=nifti_file, 
                            n_rois=n_rois, Fs=Fs,
                            subj_id=subj_id,
                            confound_strategy=confound_strategy,
                            standardize=standardize,
                            TS_name=TS_name,
                            session=session,
                    )
        else:
            BOLD_multi.concat_ts(
                nifti2timeseries(
                        nifti_file=nifti_file,
                        n_rois=n_rois, Fs=Fs,
                        subj_id=subj_id,
                        confound_strategy=confound_strategy, 
                        standardize=standardize,
                        TS_name=TS_name,
                        session=session,
                    )
            )
    return BOLD_multi


def load_TS(
        data_root, 
        file_name, 
        SESSIONs, 
        subj_id2load=None
    ):
    '''
    load a TIME_SERIES object from a .npy file
    if SESSIONs is a list, it will load all the sessions, 
        if it is a string, it will load that session
    if subj_id2load is None, it will load all the subjects
    '''
    # check if SESSIONs is a list or a string
    flag = False
    if type(SESSIONs) is str:
        SESSIONs = [SESSIONs]
        flag = True

    if subj_id2load is None:
        SUBJECTS = find_subj_list(data_root, sessions=SESSIONs)
    else:
        SUBJECTS = [subj_id2load]

    TS = {}
    for session in SESSIONs:
        TS[session] = None
        for subj in SUBJECTS:
            subj_fldr = f"{subj}_{session}"
            time_series = np.load(f"{data_root}/{subj_fldr}/{file_name}", allow_pickle='True').item()
            if TS[session] is None:
                TS[session] = time_series
            else:
                TS[session].concat_ts(time_series)

    if flag:
        return TS[SESSIONs[0]]
    return TS



####################################################################################################################################



from functions.dFC_funcs import *
import numpy as np
import time
import hdf5storage
import scipy.io as sio
import os
os.environ["MKL_NUM_THREADS"] = '64'
os.environ["NUMEXPR_NUM_THREADS"] = '64'
os.environ["OMP_NUM_THREADS"] = '64'

print('################################# CODE started running ... #################################')

################################# Parameters #################################

###### DATA PARAMETERS ######

output_root = './'

# DATA_type is either 'sample' or 'Gordon' or 'simulated' or 'ICA'
params_data_load = { 
    'DATA_type': 'Gordon', 
    'SESSIONs':['Rest1_LR' , 'Rest1_RL', 'Rest2_LR', 'Rest2_RL'], 
    'networks2include':['Auditory', 'CinguloOperc', 'Default', 'DorsalAttn', 'FrontoParietal',
        'MedialParietal', 'ParietoOccip', 'SMhand', 'SMmouth',
        'Salience', 'VentralAttn', 'Visual'], 

    'data_root_simul': './../../../../DATA/TVB data/', 
    'data_root_sample': './sampleDATA/', 
    'data_root_gordon': './../../../../DATA/HCP/HCP_Gordon/', 
    'data_root_ica': './../../../../DATA/HCP/HCP_PTN1200/node_timeseries/3T_HCP1200_MSMAll_d50_ts2/'
}

###### MEASUREMENT PARAMETERS ######

# W is in sec

params_methods = { 
    # Sliding Parameters
    'W': 44, 'n_overlap': 0.5, 'sw_method':'pear_corr', 'tapered_window':True, 
    # TIME_FREQ
    'TF_method':'WTC', 
    # CLUSTERING AND DHMM
    'clstr_base_measure':'SlidingWindow', 
    # HMM
    'hmm_iter': 30, 'dhmm_obs_state_ratio': 16/24, 
    # State Parameters
    'n_states': 12, 'n_subj_clstrs': 20, 
    # Parallelization Parameters
    'n_jobs': 2, 'verbose': 0, 'backend': 'loky', 
    # SESSION
    'session': 'Rest1_LR', 
    # Hyper Parameters
    'normalization': True, 
    'num_subj': 395, 
    'num_select_nodes': 96, 
    'num_time_point': 1200, 
    'Fs_ratio': 1.00, 
    'noise_ratio': 0.00, 
    'num_realization': 1 
}

###### HYPER PARAMETERS ALTERNATIVE ######

MEASURES_name_lst = [ 
                'SlidingWindow', 
                'Time-Freq', 
                'CAP', 
                'ContinuousHMM', 
                'Windowless', 
                'Clustering', 
                'DiscreteHMM' 
]

alter_hparams = { \
            'session': ['Rest1_RL', 'Rest2_LR', 'Rest2_RL'], 
            # 'n_overlap': [0, 0.25, 0.75, 1], 
            # 'n_states': [6, 16], 
            # # 'normalization': [], 
            # 'num_subj': [50, 100, 200], 
            # 'num_select_nodes': [30, 50, 333], 
            # 'num_time_point': [800, 1000], 
            # 'Fs_ratio': [0.50, 0.75, 1.5], 
            # 'noise_ratio': [1.00, 2.00, 3.00], 
            # 'num_realization': [] 
            }

###### dFC ANALYZER PARAMETERS ######

params_dFC_analyzer = { 
    # Parallelization Parameters
    'n_jobs': None, 'verbose': 0, 'backend': 'loky' 
}


################################# LOAD DATA #################################

data_loader = DATA_LOADER(**params_data_load)
BOLD = data_loader.load()

################################# Visualize BOLD #################################

# for session in BOLD:
#     BOLD[session].visualize(start_time=0, end_time=50, nodes_lst=list(range(10)), \
#         save_image=params_dFC_analyzer['save_image'], output_root=output_root+'BOLD_signal_'+session)

################################# Measures of dFC #################################

dFC_analyzer = DFC_ANALYZER( 
    analysis_name='reproducibility assessment', 
    **params_dFC_analyzer 
)

MEASURES_lst = dFC_analyzer.measures_initializer( 
    MEASURES_name_lst, 
    params_methods, 
    alter_hparams 
)

tic = time.time()
print('Measurement Started ...')

################################# estimate FCS #################################

task_id = int(os.getenv("SGE_TASK_ID"))
MEASURE_id = task_id-1 # SGE_TASK_ID starts from 1 not 0


if MEASURE_id >= len(MEASURES_lst):
    print("MEASURE_id out of MEASURES_lst ")
else:
    measure = MEASURES_lst[MEASURE_id]

    print("FCS estimation started...")

    time_series = BOLD[measure.params['session']]
    if measure.is_state_based:
        measure.estimate_FCS(time_series=time_series)
            
    # dFC_analyzer.estimate_group_FCS(time_series_dict=BOLD)
    print("FCS estimation done.")

    print('Measurement required %0.3f seconds.' % (time.time() - tic, ))

    # Save
    np.save(output_root+'fitted_MEASURES/MEASURE_'+str(MEASURE_id)+'.npy', measure) 
    np.save(output_root+'dFC_analyzer.npy', dFC_analyzer) 
    np.save(output_root+'data_loader.npy', data_loader) 

#################################################################################



"""
Time Series class

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import numpy as np
from requests import session
from scipy import signal
import matplotlib.pyplot as plt
from copy import deepcopy
import os

from .dfc_utils import print_dict

################################# Parameters ####################################

## visualization parameters
fig_dpi = 120
fig_bbox_inches = 'tight'
fig_pad = 0.1
show_title = False
save_fig_format = 'png'

################################# TIME_SERIES class ######################################

"""

todo:
- select nodes for visualizer
- default node list is chosen by arange !
"""

class TIME_SERIES():
    def __init__(self, data=None, subj_id=None, Fs=None, time_array=None, \
                locs=None, node_labels=None, TS_name='', session_name=''):
        
        '''
        subj_id is an id to identify the subjects
        all properties are applied to every subject separately
        for instance interval applies to TS of each subj separately

        time_array of all subjects must be equal
        '''

        assert (not data is None) and (not Fs is None) and (not subj_id is None), \
            "data, subj_id, and Fs args must be provided."
        
        assert type(locs) is np.ndarray, 'locs must be a numpy array'
        assert type(node_labels) is list, 'node_labels must be a list'
        assert locs.shape[0] == len(node_labels), 'locs and node_labels must have the same length'
        assert locs.shape[1] == 3, 'locs must have 3 columns'

        self.data_dict_ = {}
        self.data_dict_[subj_id] = {}
        self.data_dict_[subj_id]['data'] = data 
        self.data_ = None
        self.Fs_ = Fs
        self.Fs_ratio_ = 1.00
        self.noise_ratio = 0.0
        self.normalized = False
        self.TS_name_ = TS_name
        self.session_name_ = session_name
        self.n_regions_ = data.shape[0]
        self.n_time_ = data.shape[1]

        # assert self.n_regions_ < self.n_time_, \
        #     "Probably you have to transpose the time_series."

        if time_array is None:
            self.time_array_ = 1/self.Fs_ + np.arange(0, data.shape[1]/self.Fs_, 1/self.Fs_)
        else:
            self.time_array_ = time_array

        self.locs_ = locs
        self.node_labels_ = node_labels

        self.interval_ = np.arange(0, self.n_time_, dtype=int)
        self.nodes_selection_ = list(range(self.n_regions_))

    @property
    def info(self):
        print_dict(self.info_dict)

    @property
    def info_dict(self):
        info_dict = {}
        info_dict['n_time'] = self.n_time
        info_dict['n_regions'] = self.n_regions
        info_dict['Fs'] = self.Fs
        info_dict['Fs_ratio'] = self.Fs_ratio_
        info_dict['noise_ratio'] = self.noise_ratio
        info_dict['nodes_lst'] = self.nodes_lst
        info_dict['node_labels'] = self.node_labels
        info_dict['nodes_locs'] = self.locs
        info_dict['subj_id_lst'] = self.subj_id_lst
        info_dict['interval'] = self.interval
        info_dict['time'] = self.time

        return info_dict

    @property
    def data_dict(self):
        return self.data_dict_

    @property
    def data(self):
        if self.data_ is None:
            self.update_data()
        return self.data_

    def update_data(self):
        # after any change in data_dict, self.data_ is 
        # set to None and needs an update before being used
        data = None
        for subj in self.data_dict:
            if data is None:
                data = self.data_dict[subj]['data']
            else:
                data = np.concatenate((data, self.data_dict[subj]['data']), axis=1)
        self.data_ = data
        return

    @property
    def subj_id_lst(self):
        return [subj_id for subj_id in self.data_dict]

    @property
    def nodes_lst(self):
        # output shape is (n_region,) 
        return np.array(self.nodes_selection_)
        
    @property
    def interval(self):
        # output shape is (n_time,) 
        return self.interval_

    @property
    def locs(self):
        '''
        locs shape is (n_region, 3)
        '''
        if self.locs_ is None:
            return None
        else:
            return self.locs_[self.nodes_lst, :]

    @property
    def node_labels(self):
        if self.node_labels_ is None:
            return None
        else:
            return [self.node_labels_[i] for i in self.nodes_lst] 

    @property
    def Fs(self):
        return self.Fs_ * self.Fs_ratio_

    @property
    def n_time(self):
        return len(self.time)

    @property
    def n_regions(self):
        return len(self.nodes_lst)

    @property
    def time(self):
        return self.time_array_[self.interval]

    @property
    def TS_name(self):
        return self.TS_name_

    def get_subj_ts(self, subjs_id=None):
        """
        you can select time samples by their subj_id
        ! be careful about the original properties of TS hidden in new TS
        node selection will be kept.

        subjs_id is one str or a list of str
        """

        flag = 0
        if not type(subjs_id) is list:
            subjs_id = [subjs_id]
            flag = 1

        # the reason we don't make it from scratch is that we want to 
        # keep the node selection and manipulations
        new_TS = deepcopy(self)

        SUBJECTS = [subj_id for subj_id in new_TS.data_dict_]
        for subj in SUBJECTS:
            if not subj in subjs_id:
                new_TS.data_dict_.pop(subj, None)

        if flag == 1:
            new_TS.TS_name_ = self.TS_name+' subject '+subjs_id[0]

        return new_TS


    def append_ts(self, new_time_series, time_array=None, subj_id=None):
        # append new time series numpy array to existing ones
        # truncate and node selection , etc will be automatically applied to new TS;
        # However, at first the new TS must have the same properties as the original properties of 
        # the existing TSs 

        assert self.n_regions_ == new_time_series.shape[0], \
            "Number of nodes mismatch."

        assert not subj_id is None, \
            "subj_id must be provided."

        assert not subj_id in self.data_dict_, \
            "subj_id already exists."

        self.data_dict_[subj_id] = {}

        if not time_array is None:
            assert self.time_array_ == time_array, \
                'time array mismatch!'

        self.data_dict_[subj_id]['data'] = new_time_series

        self.data_ = None


    def concat_ts(self, new_TS):
        '''
        concatenate another Time Series obj 
        to the current one.
        '''
        assert self.Fs == new_TS.Fs, 'Fs mismatch!'
        assert len(self.node_labels) == len(new_TS.node_labels), 'node_labels mismatch!'
        for i in range(len(self.node_labels)):
            assert self.node_labels[i] == new_TS.node_labels[i], 'node_labels mismatch!'
            assert np.all(self.locs[i, :] == new_TS.locs[i, :]), 'locs mismatch!'

        self.append_ts(new_time_series=new_TS.data, subj_id=new_TS.subj_id_lst[0])

    def truncate(self, start_time=None, end_time=None, start_point=None, end_point=None):

        # truncates TS of every subj separately
        # based on either time or samples
        # if all None -> whole time_series
        #check if not out of total interval

        start = 0
        end = self.n_time

        if not start_point is None:
            start = start_point
        
        if not end_point is None:
            end = end_point + 1

        if not start_time is None:
            start = np.argwhere(self.time_array_>=start_time)[0,0]

        if not end_time is None:
            end = np.argwhere(self.time_array_<=end_time)[-1,0] + 1

        if start > self.interval_[0] or end < self.interval_[-1]:
            # make sure the interval is not out of range
            start = max(start, 0)
            end = min(end, self.n_time)

            self.interval_ = np.arange(start, end, dtype=int)
            
            for subj in self.data_dict_:
                self.data_dict_[subj]['data'] = self.data_dict_[subj]['data'][:, self.interval]

            self.data_ = None

    def normalize(self):
        # normalization
        if not self.normalized:
            for subj_id in self.data_dict:
                new_time_series = self.data_dict[subj_id]['data']
                for n in range(new_time_series.shape[0]):
                    new_time_series[n, :] = new_time_series[n, :] - np.mean(new_time_series[n, :])
                    new_time_series[n, :] = np.divide(new_time_series[n, :], np.std(new_time_series[n, :]))
                self.data_dict_[subj_id]['data'] = new_time_series

            self.normalized = True
            self.data_ = None

    def spatial_downsample(self, num_select_nodes, rand_node_slct=False):
        if num_select_nodes < self.n_regions:
            if rand_node_slct:
                np.random.seed(0)
                nodes_idx = np.random.choice(range(self.n_regions), size=num_select_nodes, replace=False)
                nodes_idx.sort()
            else:
                # nodes_idx = np.array(list(range(self.num_select_nodes)))
                nodes_idx = np.arange(0, self.n_regions, np.ceil(self.n_regions/num_select_nodes), dtype=int)
            self.select_nodes(nodes_idx=nodes_idx)

            self.data_ = None

    def Fs_resample(self, Fs_ratio=None):
        # downsample frequency
        if Fs_ratio != 1 and self.Fs_ratio_ == 1:
            for subj_id in self.data_dict:
                new_time_series = self.data_dict[subj_id]['data']
                downsampled_time_series = np.zeros((new_time_series.shape[0], int(new_time_series.shape[1]*Fs_ratio)))
                for n in range(new_time_series.shape[0]):
                    downsampled_time_series[n, :] =  signal.resample(new_time_series[n, :], int(new_time_series.shape[1]*Fs_ratio))
                self.data_dict_[subj_id]['data'] = downsampled_time_series
            self.Fs_ratio_ = Fs_ratio

            start_time = self.time_array_[self.interval_[0]]
            end_time = self.time_array_[self.interval_[-1]]

            _, self.time_array_ =  signal.resample(new_time_series[n, :], int(new_time_series.shape[1]*Fs_ratio), t=self.time_array_)

            start = np.argwhere(self.time_array_>=start_time)[0,0]
            end = np.argwhere(self.time_array_<=end_time)[-1,0] + 1
            self.interval_ = np.arange(start, end, dtype=int)

            self.data_ = None

    def add_noise(self, noise_ratio, mean_noise=0):
        # adding noise perturbation 
        if noise_ratio > 0 and self.noise_ratio == 0 :
            for subj_id in self.data_dict:
                new_time_series = self.data_dict[subj_id]['data']
                power_signal = np.mean(new_time_series ** 2)
                power_noise = power_signal * noise_ratio
                new_time_series += np.random.normal(mean_noise, np.sqrt(power_noise), (new_time_series.shape[0], new_time_series.shape[1]))
                self.data_dict_[subj_id]['data'] = new_time_series

            self.noise_ratio = noise_ratio
            self.data_ = None

    def select_subjs(self, num_subj):
        # selects the first num_subj subjects in self.subj_id_lst

        SUBJECTS = [subj_id for subj_id in self.data_dict_]
        if num_subj < len(SUBJECTS):
            for subj_id in SUBJECTS:
                if not subj_id in SUBJECTS[:num_subj]:
                    self.data_dict_.pop(subj_id, None)

            self.data_ = None

    def select_nodes(self, nodes_idx=None):
        # select the nodes indexed by numbers in nodes_idx. nodes_idx is a numpy 1D array
        # if nodes_idx is None -> all the nodes will be considered (resets node selection)
        # if nodes_idx is not sorted, it can be used to reorder the nodes
        # this function can be used only once (you cannot select the nodes again)

        if nodes_idx is None:
            self.nodes_selection_ = np.arange(0, self.n_regions_, 1, dtype=int)
        else:
            self.nodes_selection_ = nodes_idx  

        for subj in self.data_dict_:
            self.data_dict_[subj]['data'] = self.data_dict_[subj]['data'][self.nodes_lst, :]

        self.data_ = None

        

    def visualize(self, start_time=None, end_time=None, 
        nodes_lst=None, 
        save_image=False, output_root=None):
        '''
        time in seconds
        nodes_lst is a list of indices
        '''

        start = 0
        end = self.n_time

        if not start_time is None:
            start = np.argwhere(self.time>=start_time)[0,0]

        if not end_time is None:
            end = np.argwhere(self.time<=end_time)[-1,0] + 1
        
        interval = list(range(start, end))

        if nodes_lst is None:
            nodes_lst = self.nodes_lst
        else:
            nodes_lst = np.array(nodes_lst)[:, np.newaxis]

        plt.figure(figsize=(15, 5))
        plt.plot(self.time[interval], self.data[nodes_lst, interval].T)
        plt.xlabel('time (sec)')

        if not self.session_name_ is None:
            session_name = self.session_name_
        else:
            session_name = ''

        title = self.TS_name_ + ' ' + session_name
        if show_title:
            plt.title(title)
        if save_image:
            folder = output_root[:output_root.rfind('/')]
            file_name = title.replace(" ", "_")
            if not os.path.exists(folder):
                os.makedirs(folder)
            plt.savefig(output_root+file_name+'.'+save_fig_format, 
                dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
            ) 
            plt.close()
        else:
            plt.show()


"""
Implementation of dFC methods.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import numpy as np
import time
from sklearn.covariance import GraphicalLassoCV, graphical_lasso
from scipy import signal

from .base_dfc_method import BaseDFCMethod
from ..time_series import TIME_SERIES
from ..dfc import DFC

################################# Sliding-Window #################################

"""

Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signals.
    dt : float
        Sample spacing.

todo:
"""

class SLIDING_WINDOW(BaseDFCMethod):

    def __init__(self, **params):
        self.logs_ = ''
        self.TPM = []
        self.FCS_ = []
        self.FCS_fit_time_ = None
        self.dFC_assess_time_ = None

        self.params_name_lst = ['measure_name', 'is_state_based', 'sw_method', 'tapered_window',
            'W', 'n_overlap', 'normalization',
            'num_select_nodes', 'num_time_point', 'Fs_ratio',
            'noise_ratio', 'num_realization', 'session']
        self.params = {}
        for params_name in self.params_name_lst:
            if params_name in params:
                self.params[params_name] = params[params_name]
            else:
                self.params[params_name] = None

        self.params['measure_name'] = 'SlidingWindow'
        self.params['is_state_based'] = False

        assert self.params['sw_method'] in self.sw_methods_name_lst, \
            "sw_method not recognized."
        
    
    @property
    def measure_name(self):
        return self.params['measure_name'] #+ '_' + self.sw_method

    def shan_entropy(self, c):
        c_normalized = c / float(np.sum(c))
        c_normalized = c_normalized[np.nonzero(c_normalized)]
        H = -sum(c_normalized* np.log2(c_normalized))  
        return H

    def calc_MI(self, X, Y):
        
        bins = 20
        
        c_XY = np.histogram2d(X,Y,bins)[0]
        c_X = np.histogram(X,bins)[0]
        c_Y = np.histogram(Y,bins)[0]
        
        H_X = self.shan_entropy(c_X)
        H_Y = self.shan_entropy(c_Y)
        H_XY = self.shan_entropy(c_XY)
        
        MI = H_X + H_Y - H_XY
        return MI

    def FC(self, time_series):
    
        if self.params['sw_method']=='GraphLasso':
            model = GraphicalLassoCV()
            model.fit(time_series.T)
            C = model.covariance_
        else:
            C = np.zeros((time_series.shape[0], time_series.shape[0]))
            for i in range(time_series.shape[0]):
                for j in range(i, time_series.shape[0]):
                    
                    X = time_series[i, :]
                    Y = time_series[j, :]

                    if self.params['sw_method']=='MI':
                        ########### Mutual Information ##############
                        C[j, i] = self.calc_MI(X, Y)
                    else:
                        ########### Pearson Correlation ##############
                        if np.var(X)==0 or np.var(Y)==0:
                            C[j, i] = 0
                        else:
                            C[j, i] = np.corrcoef(X, Y)[0, 1]

                    C[i, j] = C[j, i]   
                
        return C

    def dFC(self, time_series, W=None, n_overlap=None, tapered_window=False):
        # W is in time samples
        
        L = time_series.shape[1]
        step = int((1-n_overlap)*W)
        if step == 0:
            step = 1

        window_taper = signal.windows.gaussian(W, std=3*W/22)
        # C = DFC(measure=self)
        FCSs = list()
        TR_array = list()
        for l in range(0, L-W+1, step):

            ######### creating a rectangel window ############
            window = np.zeros((L))
            window[l:l+W] = 1
            
            ########### tapering the window ##############
            if tapered_window:
                window = signal.convolve(window, window_taper, mode='same') / sum(window_taper)

            window = np.repeat(np.expand_dims(window, axis=0), time_series.shape[0], axis=0)

            # int(l-W/2):int(l+3*W/2) is the nonzero interval after tapering
            FCSs.append(self.FC( 
                        np.multiply(time_series, window)[ 
                            :,max(int(l-W/2),0):min(int(l+3*W/2),L) 
                                                        ] 
                                )
                        )
            TR_array.append(int((l + (l+W)) / 2) )

        return np.array(FCSs), np.array(TR_array)
    
    def estimate_dFC(self, time_series):
        
        '''
        we assume calc is applied on subjects separately
        '''
        assert len(time_series.subj_id_lst)==1, \
            'this function takes only one subject as input.'

        assert type(time_series) is TIME_SERIES, \
            "time_series must be of TIME_SERIES class."

        time_series = self.manipulate_time_series4dFC(time_series)

        # start timing
        tic = time.time()

        # W is converted from sec to samples
        FCSs, TR_array = self.dFC(time_series=time_series.data,
            W=int(self.params['W'] * time_series.Fs) ,
            n_overlap=self.params['n_overlap'],
            tapered_window=self.params['tapered_window']
            )

        # record time
        self.set_dFC_assess_time(time.time() - tic)

        dFC = DFC(measure=self)
        dFC.set_dFC(FCSs=FCSs, TR_array=TR_array, TS_info=time_series.info_dict)

        return dFC
    
################################################################################


"""
Implementation of dFC methods.
the parent dFC method class

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import numpy as np
from copy import deepcopy

from ..dfc_utils import SW_downsample, visualize_FCS

################################# BaseDFCMethod class ####################################

"""
todo:
- type annotation
"""

class BaseDFCMethod:

    TF_methods_name_lst = [ \
        'CWT_mag', \
        'CWT_phase_r', \
        'CWT_phase_a', \
        'WTC' \
    ]

    sw_methods_name_lst = [ \
        'pear_corr', \
        'MI', \
        'GraphLasso', \
    ]

    base_methods_name_lst = ['SlidingWindow', 'Time-Freq']

    def __init__(self):
        self.measure_name = ''
        self.is_state_based = bool()
        self._stat = []
        self.TPM = []
        self.params = {}
        self.TS_info_ = {}
        self.FCS_fit_time_ = None
        self.dFC_assess_time_ = None
        self.logs_ = ''

    @property
    def FCS_fit_time(self):
        return self.FCS_fit_time_

    @property
    def dFC_assess_time(self):
        return self.dFC_assess_time_

    @property
    def TS_info(self):
        # info of the time series used to train/estimate FCSs
        return self.TS_info_

    @property
    def is_state_based(self):
        return self.params['is_state_based']

    @property
    def FCS(self):
        return self.FCS_

    # test
    @property
    def FCS_dict(self):
        # returns a dict including FCS matrices

        if not self.is_state_based:
            return None

        C_A = self.FCS
        FCSs = {}
        for k in range(C_A.shape[0]):
            FCSs['FCS'+str(k+1)] = C_A[k,:,:]
            
        return FCSs

    @property
    def info(self):
        return self.params

    @property
    def logs(self):
        print(self.logs_)

    def issame(self, dFC):
        if type(self)==type(dFC):
            for param_name in self.params:
                if self.params[param_name] != dFC.params[param_name]:
                    return False
        else:
            return False
        return True

    #test
    def param_match(self, **param_dict):
        for param in param_dict:
            if param in self.params:
                if type(param_dict[param]) is list:
                    if not self.params[param] in param_dict[param]:
                        return False
                else:
                    if self.params[param]!=param_dict[param]:
                        return False
        return True

    def set_FCS_fit_time(self, time):
        self.FCS_fit_time_ = time

    def set_dFC_assess_time(self, time):
        self.dFC_assess_time_ = time

    def set_mean_activity(self, time_series):
        # mean activity of regions at each state
        if self.is_state_based:
            if 'sw_method' in self.params_name_lst:
                SUBJECTs = time_series.subj_id_lst
                TS_data = None
                for subject in SUBJECTs:
                    subj_TS = time_series.get_subj_ts(subjs_id=subject).data
                    new_TS_data = SW_downsample(data=subj_TS.T, \
                        Fs=time_series.Fs, W=self.params['W'], \
                        n_overlap=self.params['n_overlap'], \
                        tapered_window=self.params['tapered_window'] \
                    ).T
                    if TS_data is None:
                        TS_data = new_TS_data
                    else:
                        TS_data = np.concatenate((TS_data, new_TS_data), axis=1)
            else:
                TS_data = time_series.data
            mean_act = list()
            for i in np.unique(self.Z):
                ids = np.array([int(state==i) for state in self.Z])
                mean_act.append(np.average(TS_data, weights=ids, axis=1))
            self.mean_act = np.array(mean_act)
        else:
            self.mean_act = None

    def estimate_FCS(self, time_series=None):
        pass

    def estimate_dFC(self, time_series=None):
        pass

    def manipulate_time_series4FCS(self, time_series):
        '''
        passing None to params will not change the time series
        num_realization is not implemented yet
        '''

        new_time_series = deepcopy(time_series)

        # SUBJECTs
        if not self.params['num_subj'] is None:
            new_time_series.select_subjs(num_subj=self.params['num_subj'])
        # SPATIAL RESOLUTION
        if not self.params['num_select_nodes'] is None:
            new_time_series.spatial_downsample(num_select_nodes=self.params['num_select_nodes'], rand_node_slct=False)
        # TEMPORAL RESOLUTION
        if not self.params['Fs_ratio'] is None:
            new_time_series.Fs_resample(Fs_ratio=self.params['Fs_ratio'])
        # NORMALIZE
        if self.params['normalization']:
            new_time_series.normalize()
        # NOISE
        if not self.params['noise_ratio'] is None:
            new_time_series.add_noise(noise_ratio=self.params['noise_ratio'], mean_noise=0)
        # NUMBER OF TIME POINTS
        if not self.params['num_time_point'] is None:
            new_time_series.truncate(start_point=0, end_point=self.params['num_time_point']-1)

        self.TS_info_ = new_time_series.info_dict

        return new_time_series

    def manipulate_time_series4dFC(self, time_series):
        '''
        passing None to params will not change the time series
        num_realization is not implemented yet
        '''

        new_time_series = deepcopy(time_series)

        # SPATIAL RESOLUTION
        if not self.params['num_select_nodes'] is None:
            new_time_series.spatial_downsample(num_select_nodes=self.params['num_select_nodes'], rand_node_slct=False)
        # TEMPORAL RESOLUTION
        if not self.params['Fs_ratio'] is None:
            new_time_series.Fs_resample(Fs_ratio=self.params['Fs_ratio'])
        # NORMALIZE
        if self.params['normalization']:
            new_time_series.normalize()
        # NOISE
        if not self.params['noise_ratio'] is None:
            new_time_series.add_noise(noise_ratio=self.params['noise_ratio'], mean_noise=0)
        # NUMBER OF TIME POINTS
        if not self.params['num_time_point'] is None:
            new_time_series.truncate(start_point=0, end_point=self.params['num_time_point']-1)

        return new_time_series
    
    def visualize_states(self):
        pass

    # todo : use FCS_dict func in this func
    def visualize_FCS(
            self,
            normalize=True, fix_lim=True, 
            save_image=False, output_root=None
        ):
        
        visualize_FCS(
            self,
            normalize=normalize, fix_lim=fix_lim, 
            save_image=save_image, output_root=output_root
        )


################################## NEW METHOD ##################################

'''
by : web link

Reference: ##

Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signals.
    dt : float
        Sample spacing.

todo:

import needed_toolbox

class method_name(dFC):

    def __init__(self, **params):
        self.FCS_ = []
        self.logs_ = ''

        self.params_name_lst = ['measure_name', 'is_state_based', 'n_states',
            'normalization', 'num_subj', 'num_select_nodes', 'num_time_point',
            'Fs_ratio', 'noise_ratio', 'num_realization', 'session']
        self.params = {}
        for params_name in self.params_name_lst:
            if params_name in params:
                self.params[params_name] = params[params_name]
            else:
                self.params[params_name] = None

        self.params['specific_param'] = value
        self.params['measure_name'] = 'method_name'
        self.params['is_state_based'] = True/False
    
    @property
    def measure_name(self):
        return self.params['measure_name'] 

    def estimate_FCS(self, time_series):

        assert type(time_series) is TIME_SERIES, \
            "time_series must be of TIME_SERIES class."

        time_series = self.manipulate_time_series4FCS(time_series)

        # start timing
        tic = time.time()

        # calc FCSs

        # calc self.Z

        # mean activation of states
        self.set_mean_activity(time_series)

        # record time
        self.set_FCS_fit_time(time.time() - tic)

        return self

    def estimate_dFC(self, time_series):
        
        assert type(time_series) is TIME_SERIES, \
            "time_series must be of TIME_SERIES class."

        assert len(time_series.subj_id_lst)==1, \
            'this function takes only one subject as input.'

        time_series = self.manipulate_time_series4dFC(time_series)

        # start timing
        tic = time.time()

        # calc FCSs and FCS_idx

        # record time
        self.set_dFC_assess_time(time.time() - tic)
            
        dFC = DFC(measure=self)
        dFC.set_dFC(FCSs=self.FCS_, FCS_idx=FCS_idx, TS_info=time_series.info_dict)
        return dFC
'''



# -*- coding: utf-8 -*-
"""
Analytical functions for the comparison framework.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import numpy as np
import statsmodels.api as sm
from scipy import stats
from statsmodels.formula.api import ols
import pandas as pd
from copy import deepcopy

from ..dfc_utils import zip_name, mat_reorder, dFC_mat2vec 
 
################################# Analytical Functions ####################################

############## STAT Functions ##############

def make_sim_distribution(sim_mats_lst, name_lst, zip_names=True):
    '''
    each sim_mat in the sim_mat_lst corresponds
    to a subj. the name_lst must correspond to the
    columns and rows of sim_mats
    '''
    output = {}
    for sim_mat in sim_mats_lst:
        
        for i, name_i in enumerate(name_lst):
            for j, name_j in enumerate(name_lst):

                if j>=i:
                    continue

                if zip_names:
                    name_i_used = zip_name(name_i)
                    name_j_used = zip_name(name_j)
                else:
                    name_i_used = name_i
                    name_j_used = name_j

                if not name_i_used in output:
                    output[name_i_used] = {}
                if not name_j_used in output[name_i_used]:
                    output[name_i_used][name_j_used] = {'sim':list(), '':list()}
            
                output[name_i_used][name_j_used]['sim'].append(sim_mat[i, j])
                output[name_i_used][name_j_used][''].append('sim')
    return output

def two_way_anova(data):
    '''
    perform two-way anova
    target: sim
    factor1: session
    factor2: direction
    '''
    df = pd.DataFrame(data)

    # Performing two-way ANOVA
    model = ols('sim ~ C(session) + C(direction) +\
    C(session):C(direction)',
                data=df).fit()
    
    return sm.stats.anova_lm(model, type=2)
    
def convert_pvalue_to_asterisks(pvalue):
    if pvalue <= 0.0001:
        return "****"
    elif pvalue <= 0.001:
        return "***"
    elif pvalue <= 0.01:
        return "**"
    elif pvalue <= 0.05:
        return "*"
    return "ns"

############## Randomization Functions ##############

def randomize_time(dFC_dict, N):
    '''
    '''
    output = {}
    for n in range(N):

        for i, measure_i_name in enumerate(dFC_dict):
            
            dFC_mat_i = dFC_dict[measure_i_name]

            # randomize the temporal order
            n_time = dFC_mat_i.shape[0]
            idx = np.random.choice(n_time, n_time, replace=False)
            dFC_mat_i = dFC_mat_i[idx, :, :]

            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i)

            for j, measure_j_name in enumerate(dFC_dict):
            
                if j>i:
                    continue
                if not measure_i_name in output:
                    output[measure_i_name] = {}
                if not measure_j_name in output[measure_i_name]:
                    output[measure_i_name][measure_j_name] = {'sim':list(), '':list()}

                dFC_mat_j = dFC_dict[measure_j_name]

                # randomize the temporal order
                n_time = dFC_mat_j.shape[0]
                idx = np.random.choice(n_time, n_time, replace=False)
                dFC_mat_j = dFC_mat_j[idx, :, :]

                dFC_mat_j_vec = dFC_mat2vec(dFC_mat_j)

                sim, p = stats.spearmanr(dFC_mat_i_vec.flatten(), dFC_mat_j_vec.flatten())
                output[measure_i_name][measure_j_name]['sim'].append(sim)
                output[measure_i_name][measure_j_name][''].append('sim')

    return output


def suffle_dFC(dFC_mat, mode):
    '''
    dFC_mat = ndarray(time, region, region)
    mode can be 'temporal', 'spatial',
    or 'all'
    '''
    new_dFC_mat = deepcopy(dFC_mat)
    if mode=='temporal':
        n_time = new_dFC_mat.shape[0]
        new_order = np.random.choice(n_time, n_time, replace=False)
        new_dFC_mat = new_dFC_mat[new_order, :, :]
    elif mode=='spatial':
        n_region = new_dFC_mat.shape[1]
        new_order = np.random.choice(n_region, n_region, replace=False)
        for k, mat in enumerate(new_dFC_mat):
            new_dFC_mat[k, :, :] = mat_reorder(new_dFC_mat[k, :, :], new_order)
    elif mode=='all':
        #spatial
        n_region = new_dFC_mat.shape[1]
        new_order_regions = np.random.choice(n_region, n_region, replace=False)
        for k, mat in enumerate(new_dFC_mat):
            new_dFC_mat[k, :, :] = mat_reorder(new_dFC_mat[k, :, :], new_order_regions)
        #temporal
        n_time = new_dFC_mat.shape[0]
        new_order_time = np.random.choice(n_time, n_time, replace=False)
        new_dFC_mat = new_dFC_mat[new_order_time, :, :]

    return new_dFC_mat


def randomized_dFC_sim(dFC_dict, N, mode):
    '''
    mode can be 'temporal', 'spatial',
    or 'all'
    'spatial': this will result in different methods having
    different spatial/region orders but still the
    same temporal order
    'temporal': this will result in different methods having
    different temporal orders but still the
    same spatial order
    'all': this will result in different methods having
    different temporal orders AND different spatial order
    '''
    output = {}
    for n in range(N):

        for i, measure_i_name in enumerate(dFC_dict):
            
            dFC_mat_i = dFC_dict[measure_i_name]

            # randomize the spatial (regions) order
            dFC_mat_i = suffle_dFC(dFC_mat_i, mode=mode)

            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i)

            for j, measure_j_name in enumerate(dFC_dict):
            
                if j>i:
                    continue
                if not measure_i_name in output:
                    output[measure_i_name] = {}
                if not measure_j_name in output[measure_i_name]:
                    output[measure_i_name][measure_j_name] = {'sim':list(), '':list()}

                dFC_mat_j = dFC_dict[measure_j_name]

                # randomize the temporal order
                dFC_mat_j = suffle_dFC(dFC_mat_j, mode=mode)

                dFC_mat_j_vec = dFC_mat2vec(dFC_mat_j)

                sim, p = stats.spearmanr(dFC_mat_i_vec.flatten(), dFC_mat_j_vec.flatten())
                output[measure_i_name][measure_j_name]['sim'].append(sim)
                output[measure_i_name][measure_j_name][''].append('sim')

    return output


def dFC_rand_generator(FCS, n_time):
    '''
    generate a dFC mat of length n_time 
    using spatial FC patterns in FCS = (num_pattern, ROI, ROI)
    '''
    dFC_rand = None
    idx = np.random.choice(FCS.shape[0], n_time, replace=True)
    dFC_rand = FCS[idx, :, :]
    return dFC_rand

def dFC_rand_sim(FCS_dict, n_time, N):
    '''
    for random state TC similarity assessment
    '''
    output = {}
    for n in range(N):

        for i, measure_i_name in enumerate(FCS_dict):
                
            dFC_rand = dFC_rand_generator(FCS_dict[measure_i_name], n_time=n_time)
            dFC_mat_i = dFC_rand
            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i)

            for j, measure_j_name in enumerate(FCS_dict):
            
                if j>i:
                    continue
                if not measure_i_name in output:
                    output[measure_i_name] = {}
                if not measure_j_name in output[measure_i_name]:
                    output[measure_i_name][measure_j_name] = {'sim':list(), '':list()}

                dFC_rand = dFC_rand_generator(FCS_dict[measure_j_name], n_time=n_time)
                dFC_mat_j = dFC_rand
                dFC_mat_j_vec = dFC_mat2vec(dFC_mat_j)

                sim, p = stats.spearmanr(dFC_mat_i_vec.flatten(), dFC_mat_j_vec.flatten())
                output[measure_i_name][measure_j_name]['sim'].append(sim)
                output[measure_i_name][measure_j_name][''].append('sim')
    return output

############## Hierarchical Clustering ##############

def correct_order(s):
    list = s.split('-')
    list = [int(item) for item in list]
    list.sort()
    return '-'.join(str(x) for x in list)

def open_trees(Z, num_leaf):
    '''
    replace trees in Z by their leaves
    '''
    Z_copy = deepcopy(Z)
    Z_new = []
    for tree in Z_copy:
        Z_new.append([tree[0], tree[1], tree[2], tree[3]])
    encode_dict = {}
    counter = num_leaf
    for tree in Z_new:
        if tree[0]>=num_leaf:
            tree[0] = encode_dict[tree[0]]
        else:
            tree[0] = str(int(tree[0]))
        if tree[1]>=num_leaf:
            tree[1] = encode_dict[tree[1]]
        else:
            tree[1] = str(int(tree[1]))
        encode_dict[counter] = tree[0]+'-'+tree[1]
        encode_dict[counter] = correct_order(encode_dict[counter])
        counter += 1
    return Z_new

def is_trees_equal(trees_1, trees_2):
    '''
    trees_2 is the reference
    '''
    for tree in trees_1:
        if (not [tree[0], tree[1]] in trees_2) \
            and (not [tree[1], tree[0]] in trees_2):
            return False
    return True

def is_in_Z_clstrs(trees, Z_clstrs, trees_key):
    for key in Z_clstrs:
        if is_trees_equal(trees, Z_clstrs[key][trees_key]):
            return key
    return None

def cluster_Z(Z_lst, num_leaf):
    '''
    Z_lst is the list of linkages of samples
    num_leaf is the number of objects in clustering
    '''
    Z_clstrs = {}
    counter = 0
    for Z in Z_lst:
        # replace trees in Z by their leaves
        Z_open = open_trees(Z, num_leaf)
        trees = [[tree[0], tree[1]] for tree in Z_open]
        distances = [tree[2] for tree in Z]
        clstr_idx = is_in_Z_clstrs(trees, Z_clstrs, trees_key='trees')
        
        if clstr_idx is None:
            Z_clstrs[counter] = {
                'Z': Z, 
                'trees': trees,
                'freq': 1, 
                'distance_lst': [distances]
            }
            counter += 1
        else:
            Z_clstrs[clstr_idx]['freq'] += 1
            Z_clstrs[clstr_idx]['distance_lst'].append(distances)
    return Z_clstrs


from functions.dFC_funcs import *
import numpy as np
import time
import hdf5storage
import scipy.io as sio
import os
os.environ["MKL_NUM_THREADS"] = '64'
os.environ["NUMEXPR_NUM_THREADS"] = '64'
os.environ["OMP_NUM_THREADS"] = '64'

print('################################# subject-level similarity measurement CODE started running ... #################################')

################################# Parameters #################################
###### DATA PARAMETERS ######

input_root = './'
output_root = './'

################################# LOAD #################################

dFC_analyzer = np.load(input_root+'dFC_analyzer.npy',allow_pickle='TRUE').item()
data_loader = np.load(input_root+'data_loader.npy',allow_pickle='TRUE').item()

################################# LOAD ASSESSED dFC #################################

task_id = int(os.getenv("SGE_TASK_ID"))
subj_id = data_loader.SUBJECTS[task_id-1] # SGE_TASK_ID starts from 1 not 0
folder = input_root+'dFC_assessed/SUBJ_'+str(subj_id)

ALL_RECORDS = os.listdir(folder+'/')
ALL_RECORDS = [i for i in ALL_RECORDS if 'dFCM' in i]
ALL_RECORDS.sort()
dFCM_lst = list()
for s in ALL_RECORDS:
    dFCM = np.load(folder+'/'+s, allow_pickle='TRUE').item()
    dFCM_lst.append(dFCM)
print('assessed dFCMs loaded ...')

################################# SIMILARITY MEASUREMENT #################################

similarity_assessment = SIMILARITY_ASSESSMENT(dFCM_lst=dFCM_lst)

tic = time.time()
print('Measurement Started ...')

print("Similarity measurement started...")
SUBJ_output = similarity_assessment.run(FILTERS=dFC_analyzer.hyper_param_info, downsampling_method='default')
print("Similarity measurement done.")

print('Measurement required %0.3f seconds.' % (time.time() - tic, ))

# Save
folder = output_root+'similarity_measured'
if not os.path.exists(folder):
    os.makedirs(folder)

np.save(folder+'/SUBJ_'+str(subj_id)+'_output.npy', SUBJ_output) 
#################################################################################


from pydfc import (
    data_loader,
    MultiAnalysis,
)
import time
import numpy as np
import os
import warnings

warnings.simplefilter('ignore')

os.environ["MKL_NUM_THREADS"] = '16'
os.environ["NUMEXPR_NUM_THREADS"] = '16'
os.environ["OMP_NUM_THREADS"] = '16'

################################# Parameters #################################

# Data parameters
# main_root = '../../DATA/ds002785/' # for local
main_root = '../../../DATA/task-based/openneuro/ds002785/' # for server
dFC_assessed_root = main_root + 'dFC_assessed/'
output_root = main_root + 'validation_results/'

################################# LOAD FIT MEASURES #################################

SUBJECTS = data_loader.find_subj_list(data_root=roi_root, sessions=params_data_load['SESSIONs'])

ALL_RECORDS = os.listdir(dFC_assessed_root)
ALL_RECORDS = [i for i in ALL_RECORDS if 'dFC' in i]
ALL_RECORDS.sort()
dFC_lst = list()
for s in ALL_RECORDS:
    dFC = np.load(dFC_assessed_root+s, allow_pickle='TRUE').item()
    dFC_lst.append(dFC)
print('dFCs loaded ...')

################################# SIMILARITY MEASUREMENT #################################

# similarity_assessment = SIMILARITY_ASSESSMENT(dFCM_lst=dFCM_dict['dFCM_lst'])

# tic = time.time()
# print('Measurement Started ...')

# print("Similarity measurement started...")
# SUBJ_output = similarity_assessment.run(FILTERS=dFC_analyzer.hyper_param_info, downsampling_method='default')
# print("Similarity measurement done.")

# print('Measurement required %0.3f seconds.' % (time.time() - tic, ))

# # Save
# folder = output_root+'similarity_measured'
# if not os.path.exists(folder):
#     os.makedirs(folder)

# np.save(folder+'/SUBJ_'+str(subj_id)+'_output.npy', SUBJ_output) 

#######################################################################################